/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::ArmarXObjects::Navigator
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @author     Christian R. G. Dreher ( c dot dreher at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "Navigator.h"

#include <algorithm>
#include <cmath>
#include <cstddef>
#include <cstdint>
#include <iterator>
#include <memory>
#include <utility>

#include <Eigen/Geometry>

#include <opencv2/core/eigen.hpp>
#include <opencv2/imgcodecs.hpp>

#include <Ice/Current.h>

#include <SimoxUtility/color/ColorMap.h>
#include <SimoxUtility/color/cmaps/colormaps.h>
#include <SimoxUtility/math/convert/rpy_to_mat3f.h>
#include <VirtualRobot/MathTools.h>
#include <VirtualRobot/Robot.h>

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/exceptions/LocalException.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/logging/LogSender.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/services/tasks/PeriodicTask.h>
#include <ArmarXCore/core/services/tasks/TaskUtil.h>
#include <ArmarXCore/core/time/Clock.h>
#include <ArmarXCore/core/time/Duration.h>
#include <ArmarXCore/libraries/DecoupledSingleComponent/Decoupled.h>
#include <ArmarXCore/util/CPPUtility/trace.h>

#include "ArmarXGui/libraries/RemoteGui/Client/Widgets.h"

#include <RobotAPI/components/ArViz/Client/Client.h>
#include <RobotAPI/components/ArViz/Client/Elements.h>
#include <RobotAPI/components/ArViz/Client/elements/Color.h>
#include <RobotAPI/components/units/RobotUnit/NJointControllers/NJointControllerRegistry.h>
#include <RobotAPI/interface/ArViz/Elements.h>
#include <RobotAPI/libraries/armem/core/MemoryID.h>
#include <RobotAPI/libraries/armem/core/Time.h>
#include <RobotAPI/libraries/armem_vision/OccupancyGridHelper.h>
#include <RobotAPI/libraries/armem_vision/client/occupancy_grid/Reader.h>
#include <RobotAPI/libraries/core/remoterobot/RemoteRobot.h>

#include "armarx/navigation/server/scene_provider/SceneProvider.h"
#include <armarx/navigation/algorithms/Costmap.h>
#include <armarx/navigation/algorithms/CostmapBuilder.h>
#include <armarx/navigation/algorithms/astar/util.h>
#include <armarx/navigation/algorithms/persistence.h>
#include <armarx/navigation/algorithms/spfa/ShortestPathFasterAlgorithm.h>
#include <armarx/navigation/algorithms/util.h>
#include <armarx/navigation/client/NavigationStackConfig.h>
#include <armarx/navigation/client/PathBuilder.h>
#include <armarx/navigation/client/ice/NavigatorInterface.h>
#include <armarx/navigation/client/ice_conversions.h>
#include <armarx/navigation/components/Navigator/RemoteGui.h>
#include <armarx/navigation/conversions/eigen.h>
#include <armarx/navigation/core/types.h>
#include <armarx/navigation/factories/NavigationStackFactory.h>
#include <armarx/navigation/server/Navigator.h>
#include <armarx/navigation/server/event_publishing/MemoryPublisher.h>
#include <armarx/navigation/server/introspection/MemoryIntrospector.h>
#include <armarx/navigation/util/util.h>

namespace armarx::navigation::components
{
    ARMARX_DECOUPLED_REGISTER_COMPONENT(Navigator);
}

namespace armarx::navigation
{

    std::vector<core::Pose>
    convert(const std::vector<Eigen::Matrix4f>& wps)
    {
        using namespace armarx::navigation;
        std::vector<core::Pose> p;
        p.reserve(wps.size());
        std::transform(wps.begin(),
                       wps.end(),
                       std::back_inserter(p),
                       [](const auto& p) { return core::Pose(p); });
        return p;
    }

} // namespace armarx::navigation

namespace armarx::navigation::components
{

    components::Navigator::Navigator() : parameterizationService(nullptr, nullptr)
    // publisher(&resultsWriter, &eventsWriter)
    {
        // scene.timeServer = &timeServer;

        addPlugin(parameterizationReaderPlugin);
        addPlugin(parameterizationWriterPlugin);
        addPlugin(eventsWriterPlugin);
        addPlugin(resultsWriterPlugin);
        addPlugin(graphReaderPlugin);
        addPlugin(costmapWriterPlugin);
    
        addPlugin(virtualRobotReaderPlugin);

        parameterizationService = server::MemoryParameterizationService(
            &parameterizationReaderPlugin->get(), &parameterizationWriterPlugin->get());
    }


    components::Navigator::~Navigator() = default;

    void
    components::Navigator::onInitComponent()
    {
    }

    void
    components::Navigator::onConnectComponent()
    {
        ARMARX_TRACE;

        static bool initialized{false};

        // redirect to onReconnect to avoid any setup issues
        if (initialized)
        {
            ARMARX_INFO << "Reconnecting ...";
            onReconnectComponent();
            return;
        }


        virtualRobotReaderPlugin->get().setSyncTimeout(Duration::MilliSeconds(20));
        virtualRobotReaderPlugin->get().setSleepAfterSyncFailure(Duration::MilliSeconds(10));

        // initialize scene
        ARMARX_TRACE;
        initializeScene();

        executor.emplace(server::PlatformControllerExecutor(getControlComponentPlugin()));

        introspector = server::ArvizIntrospector(getArvizClient(), scene.robot);

        // TODO dynamic scene
        // TODO memory
        // TODO param (10)


        // robotStateUpdateTask = new PeriodicTask<Navigator>(this,
        //                                                    &Navigator::updateRobot,
        //                                                    params.robotUpdatePeriodMs,
        //                                                    false,
        //                                                    "RobotStateUpdateTask");
        // robotStateUpdateTask->start();

        navRemoteGui = std::make_unique<NavigatorRemoteGui>(remoteGui, *this);
        navRemoteGui->enable();

        initialized = true;

        ARMARX_INFO << "Initialized. Will now respond to navigation requests.";
    }

    void
    components::Navigator::onReconnectComponent()
    {
        ARMARX_TRACE;

        robotStateUpdateTask->start();

        // TODO not in all cases meaningful
        //resume();

        navRemoteGui->enable();
    }

    void
    components::Navigator::onDisconnectComponent()
    {
        ARMARX_TRACE;

        robotStateUpdateTask->stop();

        stopAll();

        navRemoteGui->disable();
    }

    void
    components::Navigator::onExitComponent()
    {
    }

    VirtualRobot::RobotPtr
    components::Navigator::getRobot()
    {
        ARMARX_TRACE;
        auto robot = virtualRobotReaderPlugin->get().getSynchronizedRobot(
            getControlComponentPlugin().getRobotUnitPlugin().getRobotUnit()->getKinematicUnit()->getRobotName(),
            VirtualRobot::RobotIO::RobotDescription::eCollisionModel,
            true);

        ARMARX_CHECK_NOT_NULL(robot);
        return robot;
    }

    void
    components::Navigator::initializeScene()
    {
        ARMARX_TRACE;

        scene.robot = getRobot();

        {
            if (not scene.staticScene.has_value())
            {
                scene.staticScene = staticScene();

                // cv::Mat1f mat;
                // cv::eigen2cv(scene.staticScene->costmap->getGrid(), mat);
                // cv::imwrite("/tmp/grid.exr", mat);
            }
        }

        // TODO dynamic scene
        {
            scene.graph = core::SceneGraph{.subgraphs = graphReaderPlugin->get().graphs()};
        }
    }

    std::string
    components::Navigator::getDefaultName() const
    {
        return "Navigator";
    }

    void
    components::Navigator::createConfig(const aron::data::dto::DictPtr& stackConfig,
                                        const std::string& callerId,
                                        const Ice::Current&)
    {
        // TODO: Not thread-safe.
        ARMARX_TRACE;
        ARMARX_INFO << "Creating config for caller '" << callerId << "'";

        parameterizationService.store(stackConfig, callerId, timeServer.now());

        server::NavigationStack stack = fac::NavigationStackFactory::create(stackConfig, scene);

        memoryIntrospectors.emplace_back(
            std::make_unique<server::MemoryIntrospector>(resultsWriterPlugin->get(), callerId));

        // keep track of all memory publishers
        memoryPublishers.emplace(
            callerId,
            std::make_unique<server::MemoryPublisher>(
                &resultsWriterPlugin->get(), &eventsWriterPlugin->get(), callerId));

        navigators.emplace(
            std::piecewise_construct,
            std::forward_as_tuple(callerId),
            std::forward_as_tuple(
                server::Navigator::Config{.stack = std::move(stack),
                                          .scene = &scene,
                                          .general = server::Navigator::Config::General{}},
                server::Navigator::InjectedServices{.executor = &executor.value(),
                                                    .publisher =
                                                        memoryPublishers.at(callerId).get(),
                                                    .introspector = &(introspector.value())}));
    }

    void
    components::Navigator::moveTo(const std::vector<Eigen::Matrix4f>& waypoints,
                                  const std::string& navigationMode,
                                  const std::string& callerId,
                                  const Ice::Current&)
    {
        ARMARX_TRACE;

        ARMARX_INFO << "moveTo() requested by caller '" << callerId << "'";
        ARMARX_CHECK(navigators.count(callerId) > 0)
            << "Navigator config for caller `" << callerId << "` not registered!";

        try
        {
            navigators.at(callerId).moveTo(convert(waypoints),
                                           core::NavigationFrameNames.from_name(navigationMode));
        }
        catch (...)
        {
            ARMARX_WARNING << GetHandledExceptionString();

            // TODO: Error handling.
            ARMARX_WARNING << "Failed to execute moveTo()."
                           << "Movement will be paused.";
            stopAll(); // TODO pause movement must not throw
            throw;
        }
    }


    void
    Navigator::moveTo2(const client::detail::Waypoints& waypoints,
                       const std::string& navigationMode,
                       const std::string& callerId,
                       const Ice::Current& c)
    {
        ARMARX_TRACE;

        ARMARX_IMPORTANT << "moveTo2 requested by caller '" << callerId << "'";

        const std::vector<client::WaypointTarget> wps = client::defrost(waypoints);

        ARMARX_CHECK(navigators.count(callerId) > 0)
            << "Navigator config for caller `" << callerId << "` not registered!";

        visualizeSPFA();
        navigators.at(callerId).moveTo(wps, core::NavigationFrameNames.from_name(navigationMode));
    }

    void
    components::Navigator::moveTowards(const Eigen::Vector3f& direction,
                                       const std::string& navigationMode,
                                       const std::string& callerId,
                                       const Ice::Current&)
    {
        // TODO: Error handling.
        ARMARX_TRACE;
        ARMARX_INFO << "MoveTowards requested by caller '" << callerId << "'";

        ARMARX_CHECK(navigators.count(callerId) > 0)
            << "Navigator config for caller `" << callerId << "` not registered!";

        navigators.at(callerId).moveTowards(direction,
                                            core::NavigationFrameNames.from_name(navigationMode));
    }

    void
    components::Navigator::pause(const std::string& configId, const Ice::Current&)
    {
        ARMARX_CHECK(navigators.count(configId) > 0)
            << "Navigator config for caller `" << configId << "` not registered!";
        navigators.at(configId).pause();
    }

    void
    components::Navigator::resume(const std::string& configId, const Ice::Current&)
    {
        ARMARX_TRACE;
        ARMARX_CHECK(navigators.count(configId) > 0)
            << "Navigator config for caller `" << configId << "` not registered!";
        navigators.at(configId).resume();
    }

    void
    components::Navigator::stop(const std::string& configId, const Ice::Current&)
    {
        // FIXME make sure that event is emitted
        ARMARX_TRACE;
        ARMARX_CHECK(navigators.count(configId) > 0)
            << "Navigator config for caller `" << configId << "` not registered!";
        navigators.at(configId).stop();

        // emit UserAbortTriggered event
        const core::UserAbortTriggeredEvent event{{timeServer.now()},
                                                  core::Pose(scene.robot->getGlobalPose())};
        memoryPublishers.at(configId)->userAbortTriggered(event);
    }

    void
    components::Navigator::stopAll(const Ice::Current&)
    {
        ARMARX_IMPORTANT << "stopAll()";

        ARMARX_TRACE;
        for (auto& [_, navigator] : navigators)
        {
            navigator.stop();
        }

        // emit UserAbortTriggered event
        const core::UserAbortTriggeredEvent event{{timeServer.now()},
                                                  core::Pose(scene.robot->getGlobalPose())};
        for (auto& [callerId, memoryPublisher] : memoryPublishers)
        {
            memoryPublisher->userAbortTriggered(event);
        }
    }

    bool
    components::Navigator::isPaused(const std::string& configId, const Ice::Current&)
    {
        ARMARX_TRACE;
        ARMARX_CHECK(navigators.count(configId) > 0)
            << "Navigator config for caller `" << configId << "` not registered!";
        return navigators.at(configId).isPaused();
    }

    bool
    components::Navigator::isStopped(const std::string& configId, const Ice::Current&)
    {
        ARMARX_TRACE;
        ARMARX_CHECK(navigators.count(configId) > 0)
            << "Navigator config for caller `" << configId << "` not registered!";
        return navigators.at(configId).isStopped();
    }

    armarx::PropertyDefinitionsPtr
    components::Navigator::createPropertyDefinitions()
    {
        ARMARX_TRACE;

        PropertyDefinitionsPtr def = new ComponentPropertyDefinitions(getConfigIdentifier());

        // Publish to a topic (passing the TopicListenerPrx).
        // def->topic(myTopicListener);

        // Subscribe to a topic (passing the topic name).
        // def->topic<PlatformUnitListener>("MyTopic");

        // Use (and depend on) another component (passing the ComponentInterfacePrx).
        def->component(remoteGui, "RemoteGuiProvider");
        // Add a required property.
        // def->required(properties.boxLayerName, "p.box.LayerName", "Name of the box layer in ArViz.");

        // Add an optionalproperty.
        def->optional(params.occupiedGridThreshold,
                      "p.occupancy_grid.occopied_threshold",
                      "Threshold for each cell to be considered occupied. Increase this value to "
                      "reduce noise.");

        return def;
    }

    void
    visualize(const algorithms::Costmap& costmap, viz::Client& arviz, const std::string& name)
    {
        const auto cmap = simox::color::cmaps::viridis();
        const float vmax = costmap.getGrid().array().maxCoeff();

        const auto asColor = [&cmap, &vmax](const float distance) -> viz::data::Color
        {
            const auto color = cmap.at(distance, 0.F, vmax);
            return {color.a, color.r, color.g, color.b};
        };

        auto layer = arviz.layer("costmap_" + name);

        const std::int64_t cols = costmap.getGrid().cols();
        const std::int64_t rows = costmap.getGrid().rows();

        auto mesh = viz::Mesh("");

        std::vector<std::vector<Eigen::Vector3f>> vertices;
        std::vector<std::vector<viz::data::Color>> colors;

        for (int r = 0; r < rows; r++)
        {
            auto& verticesRow = vertices.emplace_back(cols);
            auto& colorsRow = colors.emplace_back(cols);

            for (int c = 0; c < cols; c++)
            {
                verticesRow.at(c) = conv::to3D(costmap.toPositionGlobal({r, c}));
                colorsRow.at(c) = asColor(costmap.getGrid()(r, c));
            }
        }

        mesh.grid2D(vertices, colors);

        layer.add(mesh);

        arviz.commit(layer);
    }

    core::StaticScene
    components::Navigator::staticScene()
    {
        ARMARX_TRACE;

        const objpose::ObjectPoseSeq objectPoses = ObjectPoseClientPluginUser::getObjectPoses();

        // remove those objects that belong to an object dataset. the manipulation object / distance computation is broken
        const auto objectPosesStatic =
            armarx::navigation::util::filterObjects(objectPoses, {"KIT", "HOPE", "MDB", "YCB"});

        const auto objects = armarx::navigation::util::asSceneObjects(objectPosesStatic);
        ARMARX_INFO << objects->getSize() << " objects in the scene";

        ARMARX_INFO << "Creating costmap";
        ARMARX_CHECK_NOT_NULL(scene.robot);

        algorithms::CostmapBuilder costmapBuilder(
            scene.robot,
            objects,
            algorithms::Costmap::Parameters{.binaryGrid = false, .cellSize = 100},
            "Platform-navigation-colmodel");

        const auto costmap = costmapBuilder.create();

        ARMARX_INFO << "Storing costmap in memory";
        costmapWriterPlugin->get().store(
            costmap, "distance_to_obstacles", getName(), armarx::Clock::Now());

        ARMARX_INFO << "Done";

        ARMARX_TRACE;
        const auto grid = costmap.getGrid();

        core::StaticScene scene{.objects = objects,
                                .costmap = std::make_unique<algorithms::Costmap>(costmap)};

        ARMARX_INFO << "The object scene consists of " << scene.objects->getSize() << " objects";


        visualize(costmap, arviz, "distance");

        /*
        const armem::vision::occupancy_grid::client::Reader::Result result =
            occupancyGridReader.query(armem::vision::occupancy_grid::client::Reader::Query{
                .providerName = "CartographerMappingAndLocalization",
                .timestamp = armem::Time::Now()});

        if (result and result.occupancyGrid.has_value() and false) // Feature disabled on master ...
        {
            ARMARX_INFO << "Occupancy grid available!";

            const auto occupancyGridSceneElements = util::asSceneObjects(
                result.occupancyGrid.value(),
                OccupancyGridHelper::Params{.freespaceThreshold = 0.45F,
                                            .occupiedThreshold = params.occupiedGridThreshold});
            ARMARX_INFO << occupancyGridSceneElements->getSize()
                        << " scene elements from occupancy grid";

            scene.objects->addSceneObjects(occupancyGridSceneElements);

            // draw
            auto layer = arviz.layer("occupancy_grid");

            for (const auto& sceneObject : occupancyGridSceneElements->getSceneObjects())
            {
                const Eigen::Isometry3f world_T_obj(sceneObject->getGlobalPose());
                ARMARX_INFO << world_T_obj.translation();
                ARMARX_INFO << layer.size();
                layer.add(viz::Box("box_" + std::to_string(layer.size()))
                              .pose(world_T_obj)
                              .size(result.occupancyGrid->resolution)
                              .color(viz::Color::orange()));
            }

            ARMARX_INFO << "Creating costmap";

            algorithms::CostmapBuilder costmapBuilder(
                getRobot(),
                scene.objects,
                algorithms::Costmap::Parameters{.binaryGrid = false, .cellSize = 100},
                "Platform-navigation-colmodel");

            const auto costmap = costmapBuilder.create();

            ARMARX_INFO << "Done";

            ARMARX_TRACE;
            ARMARX_INFO << "Saving costmap.";
            algorithms::save(costmap, "/tmp/navigation-costmap");

            arviz.commit({layer});
        }
        else
        {
            ARMARX_INFO << "Occupancy grid not available";
        }*/

        return scene;
    }

    void
    components::Navigator::visualizeSPFA()
    {
        algorithms::spfa::ShortestPathFasterAlgorithm::Parameters spfaParams{
            .obstacleDistanceCosts = true};

        ARMARX_INFO << "spfa...";
        algorithms::spfa::ShortestPathFasterAlgorithm spfa(*scene.staticScene->costmap, spfaParams);
        const auto result123 = spfa.spfa(getRobot()->getGlobalPosition().head<2>());
        ARMARX_INFO << "spfa done...";

        const auto sceneBounds = algorithms::computeSceneBounds(scene.staticScene->objects);

        ARMARX_INFO << "Costmap SPFA";
        algorithms::Costmap cm(result123.distances, algorithms::Costmap::Parameters{}, sceneBounds);
        visualize(cm, arviz, "spfa");
        ARMARX_INFO << "Done.";
    }


    void
    components::Navigator::updateRobot()
    {
        const auto timestamp = armem::Time::Now();

        auto& virtualRobotReader = virtualRobotReaderPlugin->get();

        ARMARX_TRACE;
        ARMARX_CHECK(virtualRobotReader.synchronizeRobot(*scene.robot, timestamp));

        ARMARX_TRACE;
        const auto robotDescription =
            virtualRobotReader.queryDescription(scene.robot->getType(), timestamp);
        ARMARX_CHECK(robotDescription.has_value());

        // synchronizing platform state
        ARMARX_TRACE;
        const std::optional<armem::robot::PlatformState> platformState =
            virtualRobotReader.queryPlatformState(robotDescription.value(), armem::Time::Now());
        ARMARX_CHECK(platformState.has_value());

        // TODO / FIXME make this thread safe!
        {
            if (not scene.platformVelocity.has_value())
            {
                scene.platformVelocity = core::Twist();
            }

            scene.platformVelocity->linear = platformState->twist.linear;
            scene.platformVelocity->angular = platformState->twist.angular;
        }
    }

    server::Navigator*
    components::Navigator::activeNavigator()
    {
        ARMARX_TRACE;
        // We define the active navigator to be the one whose movement is enabled.
        const auto isActive = [](auto& nameNavPair) -> bool
        {
            server::Navigator& navigator = nameNavPair.second;
            return not navigator.isPaused();
        };

        auto it = std::find_if(navigators.begin(), navigators.end(), isActive);

        // no navigator active?
        if (it == navigators.end())
        {
            return nullptr;
        }

        return &it->second;
    }

} // namespace armarx::navigation::components
